from enum import IntEnum
from typing import Any, Tuple

import trio

from .openrover_data import MOTOR_EFFORT_FORMAT, OPENROVER_DATA_ELEMENTS
from .serial_trio import SerialTrio
from .util import OpenRoverException

SERIAL_START_BYTE = bytes([253])


class CommandVerbs(IntEnum):
    NOP = 0
    GET_DATA = 10
    SET_FAN_SPEED = 20
    RESTART = 230
    SET_DRIVE_MODE = 240
    FLIPPER_CALIBRATE = 250


def encode_packet(*args: bytes):
    payload = b''.join(args)
    return SERIAL_START_BYTE + payload + bytes([checksum(payload)])


def checksum(values):
    return 255 - sum(values) % 255


class OpenRoverProtocol:
    def __init__(self, serial: SerialTrio):
        """Low-level communication for OpenRover"""
        self._serial = serial
        # A packet involves multiple read operations, so we must lock the device for reading
        self._read_lock = trio.StrictFIFOLock()

    async def read_one_raw(self) -> bytes:
        async with self._read_lock:
            _ = await self._serial.read_until(SERIAL_START_BYTE)
            packet = SERIAL_START_BYTE + await self._serial.read_exactly(4)
            assert len(packet) == 5
            if packet[4] == checksum(packet[1:4]):
                return packet[1:4]
            else:
                raise OpenRoverException('Bad checksum. Discarding data ' + repr(packet))

    async def read_one(self) -> Tuple[int, Any]:
        """Get a strongly typed version of the data"""
        raw_data = await self.read_one_raw()
        data_element_index = raw_data[0]
        element_descriptor = OPENROVER_DATA_ELEMENTS[data_element_index]
        data_element_value = element_descriptor.data_format.unpack(raw_data[1:])
        return data_element_index, data_element_value

    async def write(self, motor_left: float, motor_right: float, flipper: float, command_verb: CommandVerbs, command_arg: int):
        binary = encode_packet(MOTOR_EFFORT_FORMAT.pack(motor_left),
                               MOTOR_EFFORT_FORMAT.pack(motor_right),
                               MOTOR_EFFORT_FORMAT.pack(flipper),
                               bytes([command_verb, command_arg]))
        await self._serial.write(binary)



import abc
from dataclasses import dataclass
from enum import IntEnum
import typing

from .util import OpenRoverException


class ReadDataFormat(abc.ABC):
    python_type: typing.Type = None

    @abc.abstractmethod
    def description(self):
        raise NotImplementedError

    @abc.abstractmethod
    def unpack(self, b: bytes):
        raise NotImplementedError


class WriteDataFormat(abc.ABC):
    python_type: typing.Type = None

    @abc.abstractmethod
    def description(self):
        raise NotImplementedError

    @abc.abstractmethod
    def pack(self, value) -> bytes:
        raise NotImplementedError


class IntDataFormat(ReadDataFormat, WriteDataFormat):
    def __init__(self, nbytes, signed):
        self.nbytes = nbytes
        self.signed = signed

    def description(self):
        s = 'signed' if self.signed else 'unsigned'
        n = self.nbytes * 8
        return f'{s} integer ({n} bits)'

    def pack(self, value):
        try:
            return int(value).to_bytes(self.nbytes, byteorder='big', signed=self.signed)
        except Exception as e:
            raise

    def unpack(self, b: bytes):
        return int.from_bytes(b, byteorder='big', signed=self.signed)


OPENROVER_LEGACY_VERSION = 16421


@dataclass
class OpenRoverFirmwareVersion:
    value: int

    def __post_init__(self):
        if self.value == 0:
            raise OpenRoverException('invalid version number %s', self.value)

    @property
    def major(self):
        if self.value == OPENROVER_LEGACY_VERSION:
            return 0
        return self.value // 10000

    @property
    def minor(self):
        if self.value == OPENROVER_LEGACY_VERSION:
            return 0
        return self.value // 100 % 100

    @property
    def patch(self):
        if self.value == OPENROVER_LEGACY_VERSION:
            return 0
        return self.value % 10


class DataFormatFirmwareVersion(ReadDataFormat):
    python_type = OpenRoverFirmwareVersion

    def unpack(self, b):
        v = UINT16.unpack(b)
        return OpenRoverFirmwareVersion(v)

    def description(self):
        return 'XYYZZ, where X=major version, Y=minor version, Z = patch version. e.g. 10502 = version 1.05.02. The special value 16421 represents pre-1.3 versions'


class DataFormatChargerState(ReadDataFormat, WriteDataFormat):
    CHARGER_ACTIVE_MAGIC_BYTES = bytes.fromhex('dada')
    CHARGER_INACTIVE_MAGIC_BYTES = bytes.fromhex('0000')
    python_type = bool

    def pack(self, value):
        if value:
            return self.CHARGER_ACTIVE_MAGIC_BYTES
        else:
            return self.CHARGER_INACTIVE_MAGIC_BYTES

    def unpack(self, b):
        return (bytes(b) == self.CHARGER_ACTIVE_MAGIC_BYTES)

    def description(self):
        return '0xDADA if charging, else 0x0000'


@dataclass
class BatteryStatus:
    overcharged_alarm: bool
    terminate_charge_alarm: bool
    over_temp_alarm: bool
    terminate_discharge_alarm: bool
    remaining_capacity_alarm: bool
    remaining_time_alarm: bool
    initialized: bool
    discharging: bool
    fully_charged: bool
    fully_discharged: bool


class DataFormatBatteryStatus(ReadDataFormat):
    python_type = BatteryStatus

    def unpack(self, b: bytes):
        assert len(b) == 2
        as_int = int.from_bytes(b, byteorder='big', signed=False)
        return BatteryStatus(
            overcharged_alarm=bool(as_int & 0x8000),
            terminate_charge_alarm=bool(as_int & 0x4000),
            over_temp_alarm=bool(as_int & 0x1000),
            terminate_discharge_alarm=bool(as_int & 0x0800),
            remaining_capacity_alarm=bool(as_int & 0x0200),
            remaining_time_alarm=bool(as_int & 0x0100),
            initialized=bool(as_int & 0x0080),
            discharging=bool(as_int & 0x0040),
            fully_charged=bool(as_int & 0x0020),
            fully_discharged=bool(as_int & 0x0010)
        )

    def description(self):
        return 'bit flags'


class DriveMode(IntEnum):
    OPEN_LOOP = 0
    CLOSED_LOOP = 1


UINT16 = IntDataFormat(2, False)
INT16 = IntDataFormat(2, True)
UINT8 = IntDataFormat(1, signed=False)


class DataFormatFixedPrecision(ReadDataFormat, WriteDataFormat):
    """A fractional number packed as an integer, but representing a fractional number"""

    def __init__(self, base_type, step=1.0, zero=0.0):
        self.base_type = base_type
        # a change of 1 in the python type corresponds to a change of this many in the base type
        self.step = step
        # the value of 0 in the python type corresponds to this value in the base type
        self.zero = zero

    def unpack(self, b: bytes):
        n = self.base_type.unpack(b)
        return (n - self.zero) / self.step

    def pack(self, p):
        n = round(p * self.step + self.zero)
        return self.base_type.pack(n)

    def description(self):
        return f'fractional (resolution=1/{self.step}, zero={self.zero}) stored as {self.base_type.description()}'


class DataFormatDriveMode(ReadDataFormat):
    python_type = DriveMode

    def unpack(self, b: bytes):
        return DriveMode(UINT16.unpack(b))

    def pack(self, p: DriveMode):
        return UINT16.pack(p.value)

    def description(self):
        return DriveMode.__doc__


OLD_CURRENT_FORMAT = DataFormatFixedPrecision(UINT16, 34)

SIGNED_MILLIS_FORMAT = DataFormatFixedPrecision(INT16, 1000)
UNSIGNED_MILLIS_FORMAT = DataFormatFixedPrecision(UINT16, 1000)
OLD_VOLTAGE_FORMAT = DataFormatFixedPrecision(UINT16, 58)
FAN_SPEED_RESPONSE_FORMAT = DataFormatFixedPrecision(UINT16, 240)
DECIKELVIN_FORMAT = DataFormatFixedPrecision(UINT16, 10, zero=2731.5)
PERCENTAGE_FORMAT = DataFormatFixedPrecision(UINT16, 100)
MOTOR_EFFORT_FORMAT = DataFormatFixedPrecision(UINT8, 125, 125)
CHARGER_STATE_FORMAT = DataFormatChargerState()
FIRMWARE_VERSION_FORMAT = DataFormatFirmwareVersion()
DRIVE_MODE_FORMAT = DataFormatDriveMode()
BATTERY_STATUS_FORMAT = DataFormatBatteryStatus()


@dataclass
class DataElement:
    index: int
    data_format: ReadDataFormat
    name: str
    description: str = None
    not_implemented: bool = False


elements = [
    DataElement(0, OLD_CURRENT_FORMAT, 'battery (A+B) current (external)', 'total current from batteries'),
    DataElement(2, UINT16, 'left motor speed', not_implemented=True),
    DataElement(4, UINT16, 'right motor speed', not_implemented=True),
    DataElement(6, UINT16, 'flipper position 1', 'flipper position sensor 1. 0=15 degrees; 1024=330 degrees;'),
    DataElement(8, UINT16, 'flipper position 2', 'flipper position sensor 2. 0=15 degrees; 1024=330 degrees;'),
    DataElement(10, OLD_CURRENT_FORMAT, 'left motor current'),
    DataElement(12, OLD_CURRENT_FORMAT, 'right motor current'),
    DataElement(14, UINT16, 'left motor encoder count', 'May overflow or underflow. Increments when motor driven forward, decrements backward'),
    DataElement(16, UINT16, 'right motor encoder count', 'May overflow or underflow. Increments when motor driven forward, decrements backward'),
    DataElement(18, UINT16, 'motors fault flag'),
    DataElement(20, UINT16, 'left motor temperature'),
    DataElement(22, UINT16, 'right motor temperature', not_implemented=True),
    DataElement(24, OLD_VOLTAGE_FORMAT, 'battery A voltage (external)'),
    DataElement(26, OLD_VOLTAGE_FORMAT, 'battery B voltage (external)'),
    DataElement(28, UINT16, 'left motor encoder interval', '0 when motor stopped. Else proportional to motor period (inverse motor speed)'),
    DataElement(30, UINT16, 'right motor encoder interval', '0 when motor stopped. Else proportional to motor period (inverse motor speed)'),
    DataElement(32, UINT16, 'flipper motor encoder interval', '0 when motor stopped. Else proportional to motor period (inverse motor speed)', not_implemented=True),
    DataElement(34, PERCENTAGE_FORMAT, 'battery A state of charge', 'Proportional charge, 0.0=empty, 1.0=full'),
    DataElement(36, PERCENTAGE_FORMAT, 'battery B state of charge', 'Proportional charge, 0.0=empty, 1.0=full'),
    DataElement(38, CHARGER_STATE_FORMAT, 'battery charging state'),
    DataElement(40, FIRMWARE_VERSION_FORMAT, 'release version'),
    DataElement(42, SIGNED_MILLIS_FORMAT, 'battery A current (external)'),
    DataElement(44, SIGNED_MILLIS_FORMAT, 'battery B current (external)'),
    DataElement(46, UINT16, 'motor flipper angle'),
    DataElement(48, FAN_SPEED_RESPONSE_FORMAT, 'fan speed'),
    DataElement(50, DRIVE_MODE_FORMAT, 'drive mode'),
    DataElement(52, BATTERY_STATUS_FORMAT, 'battery A status'),
    DataElement(54, BATTERY_STATUS_FORMAT, 'battery B status'),
    DataElement(56, UINT16, 'battery A mode'),
    DataElement(58, UINT16, 'battery B mode'),
    DataElement(60, DECIKELVIN_FORMAT, 'battery A temperature (internal)'),
    DataElement(62, DECIKELVIN_FORMAT, 'battery B temperature (internal)'),
    DataElement(64, UNSIGNED_MILLIS_FORMAT, 'battery A voltage (internal)'),
    DataElement(66, UNSIGNED_MILLIS_FORMAT, 'battery B voltage (internal)'),
    DataElement(68, SIGNED_MILLIS_FORMAT, 'battery A current (internal)', '>0 = charging; <0 = discharging'),
    DataElement(70, SIGNED_MILLIS_FORMAT, 'battery B current (internal)', '>0 = charging; <0 = discharging'),
]

OPENROVER_DATA_ELEMENTS = {e.index: e for e in elements}


def strike(s):
    return f'~~{s}~~'


def doc():
    lines = []
    lines.append('| # | Name | Data Type | Description |')
    lines.append('| - | ---- | --------- | ----------- |')

    for de in elements:
        lines.append(f'| {strike(de.index) if de.not_implemented else de.index} | {de.name} | {de.data_format.description()} | {de.description or ""} |')
    return '\n'.join(lines)


if __name__ == '__main__':
    print(doc())


def fix_encoder_delta(delta):
    MAX_ENCODER = 2 ** 16
    delta %= MAX_ENCODER
    if delta < MAX_ENCODER / 2:
        return delta
    else:
        return delta - MAX_ENCODER



import os
from pathlib import Path
import shutil
import subprocess
from subprocess import list2cmdline

import pytest
import trio

import openrover
from openrover.find_device import open_rover_device
from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.openrover_protocol import CommandVerbs, OpenRoverProtocol
from openrover.util import RoverDeviceNotFound


@pytest.fixture
async def device():
    try:
        async with open_rover_device() as dev:
            yield dev
    except RoverDeviceNotFound:
        pytest.skip('No openrover device found')


@pytest.fixture
def powerboard_firmware_file():
    p = Path(openrover.__path__[0], 'tests/resources/PowerBoard-1.5.0.hex')
    assert p.is_file()
    return p


@pytest.fixture
def booty_exe():
    maybe_booty = shutil.which('booty')
    if maybe_booty is None:
        pytest.skip('Could not test bootloader. Booty executable does not exist or is not in the executable path.')
    p = Path(maybe_booty)
    assert p.is_file()
    return p


async def test_reboot(device):
    orp = OpenRoverProtocol(device)
    # set a long timeout in case rover is already in bootloader

    try:
        # reboot the device
        await orp.write(0, 0, 0, CommandVerbs.RESTART, 0)

        # check that the device did turn off
        await orp.write(0, 0, 0, CommandVerbs.GET_DATA, 40)
        with pytest.raises(trio.TooSlowError):
            with trio.fail_after(1):
                await orp.read_one()

        # check that the device comes back up and starts responding to request for version
        with trio.fail_after(30):
            while True:
                with trio.move_on_after(1):
                    await orp.write(0, 0, 0, CommandVerbs.GET_DATA, 40)
                    k, v = await orp.read_one()
                    if k == 40:
                        return
    except:
        await trio.sleep(30)
        raise


def mark_dangerous_test(reason):
    VARNAME = 'PYTEST_DANGER_OKAY'
    VARVALUE = '1'
    return pytest.mark.skipif(os.getenv(VARNAME) != VARVALUE, reason=reason + f'\nSet the environment variable {VARNAME}={VARVALUE} to run anyway.')


@mark_dangerous_test('This test will wipe your firmware and will take a long time.')
async def test_bootloader(powerboard_firmware_file, booty_exe):
    async with open_rover_device() as device:
        orp = OpenRoverProtocol(device)
        await orp.write(0, 0, 0, CommandVerbs.RESTART, 0)
        port = device.port
        # flash rover firmware
        args = [
            str(booty_exe),
            '--port', str(port),
            '--baudrate', '57600',
            '--hexfile', str(powerboard_firmware_file.absolute()),
            '--erase',
            '--load',
            '--verify'
        ]

    print('running bootloader: ' + list2cmdline(args))

    with trio.fail_after(60 * 15):
        async with trio.Process(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as booty:
            async with trio.open_nursery() as nursery:
                async def check_stdout():
                    line_generator = stream_to_lines(booty.stdout)
                    lines = []
                    try:
                        while True:
                            with trio.fail_after(30):
                                a_line = await line_generator.__anext__()
                            lines.append(a_line)
                    except trio.TooSlowError:
                        pytest.fail(f'booty became unresponsive after output {lines}')
                    except StopAsyncIteration:
                        pass
                    await trio.sleep(1)
                    assert any(['device verified' in a_line for a_line in lines])

                async def check_stderr():
                    error_output = await stream_to_string(booty.stderr)
                    if 'device not responding' in error_output:
                        pytest.fail('Rover did not respond to booty. Does it have a bootloader?')
                    assert error_output.strip() == ''

                async def check_retcode():
                    assert await booty.wait() == 0, 'Booty exited with nonzero return code'

                nursery.start_soon(check_stderr)
                nursery.start_soon(check_stdout)
                nursery.start_soon(check_retcode)

    with trio.fail_after(30):
        while True:
            with trio.move_on_after(1):
                await orp.write(0, 0, 0, CommandVerbs.GET_DATA, 40)
                k, version = await orp.read_one()
                assert k == 40
                assert isinstance(version, OpenRoverFirmwareVersion)
                assert (version.major, version.minor, version.patch) == (1, 5, 0)


async def stream_to_string(stream: trio.abc.ReceiveStream):
    buf = ''
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
    return buf


async def stream_to_lines(stream: trio.abc.ReceiveStream):
    buf = ''
    async with stream:
        while True:
            new_bytes = await stream.receive_some(1)
            if not new_bytes:
                break
            buf += new_bytes.decode()
            *some_lines, buf = buf.splitlines()
            for line in some_lines:
                yield line
        if buf != '':
            yield buf



from openrover.openrover_data import FAN_SPEED_RESPONSE_FORMAT, MOTOR_EFFORT_FORMAT


def test_fan_speed_format():
    assert list(FAN_SPEED_RESPONSE_FORMAT.pack(0)) == [0, 0]
    assert list(FAN_SPEED_RESPONSE_FORMAT.pack(0.5)) == [0, 120]
    assert list(FAN_SPEED_RESPONSE_FORMAT.pack(1)) == [0, 240]


def test_motor_effort():
    assert list(MOTOR_EFFORT_FORMAT.pack(-1)) == [0]
    assert list(MOTOR_EFFORT_FORMAT.pack(0)) == [125]
    assert list(MOTOR_EFFORT_FORMAT.pack(+1)) == [250]



import trio
from typing import AsyncIterable


async def alist(aiterable: AsyncIterable):
    result = []
    async for d in aiterable:
        result.append(d)
    return result


async def timeout_each(delay: float, aiterable: AsyncIterable):
    assert delay >= 0
    iter = aiterable.__aiter__()
    while True:
        with trio.fail_after(delay):
            yield await  iter.__anext__()


async def limit(limit: int, aiterable: AsyncIterable):
    assert limit >= 0
    iter = aiterable.__aiter__()
    try:
        for i in range(limit):
            yield await iter.__anext__()
    except StopAsyncIteration:
        raise


async def timeout_all(delay: float, aiterable: AsyncIterable):
    assert delay >= 0
    with trio.fail_after(delay):
        async for i in aiterable:
            yield i


import logging
from typing import Any, Dict, Iterable, Mapping, Optional, Tuple

from async_generator import asynccontextmanager
import trio

from openrover.find_device import open_rover_device
from openrover.openrover_data import OPENROVER_DATA_ELEMENTS
from .openrover_protocol import CommandVerbs, OpenRoverProtocol
from .serial_trio import DeviceClosedException, SerialTrio
from .util import OpenRoverException


@asynccontextmanager
async def open_rover(path_to_serial: Optional[str] = None):
    async with trio.open_nursery() as nursery:
        if path_to_serial is None:
            device_cxt = open_rover_device()
        else:
            device_cxt = SerialTrio(path_to_serial)

        async with device_cxt as device:
            rover = Rover(nursery)
            await rover.set_device(device)
            yield rover


class Rover:
    _motor_left = 0
    _motor_right = 0
    _motor_flipper = 0

    _nursery = None
    _rover_protocol = None
    _openrover_data_to_memory_channel: Mapping[int, Tuple[trio.abc.SendChannel, trio.abc.ReceiveChannel]]

    def __init__(self, nursery):
        """An OpenRover object"""
        self._motor_left = 0
        self._motor_right = 0
        self._motor_flipper = 0
        self._nursery = nursery
        self._openrover_data_to_memory_channel = {i: trio.open_memory_channel(0) for i in OPENROVER_DATA_ELEMENTS.keys()}

    async def set_device(self, device: SerialTrio):
        self._device = device
        self._rover_protocol = OpenRoverProtocol(device)
        self._nursery.start_soon(self.process_messages)

    async def process_messages(self):
        try:
            while True:
                k, v = await self._rover_protocol.read_one()
                try:
                    snd, rcv = self._openrover_data_to_memory_channel[k]
                    snd.send_nowait(v)
                except Exception as e:
                    logging.warning(f"could not handle incoming data {k}:{v} {e}")
        except DeviceClosedException:
            pass

    def set_motor_speeds(self, left, right, flipper):
        assert -1 <= left <= 1
        assert -1 <= right <= 1
        assert -1 <= flipper <= 1
        self._motor_left = left
        self._motor_right = right
        self._motor_flipper = flipper

    async def _send_command(self, cmd, arg):
        await self._rover_protocol.write(self._motor_left, self._motor_right, self._motor_flipper, cmd, arg)

    async def send_speed(self):
        await self._send_command(CommandVerbs.NOP, 0)

    async def set_fan_speed(self, fan_speed):
        assert 0 <= fan_speed <= 1
        await self._send_command(CommandVerbs.SET_FAN_SPEED, int(fan_speed * 240))

    async def flipper_calibrate(self):
        await self._send_command(CommandVerbs.FLIPPER_CALIBRATE, int(CommandVerbs.FLIPPER_CALIBRATE))

    async def get_data(self, index) -> Any:
        """Get the next value for the given data index.
        The type of the returned value depends on the index passed."""
        await self._send_command(CommandVerbs.GET_DATA, index)
        send_channel, rcv_channel = self._openrover_data_to_memory_channel[index]
        return await rcv_channel.receive()

    async def get_data_items(self, indices: Iterable[int]) -> Dict[int, Any]:
        return {i: await self.get_data(i) for i in set(indices)}


async def get_openrover_version(port):
    try:
        async with trio.fail_after(1):
            async with SerialTrio(port) as device:
                orp = OpenRoverProtocol(device)
                while True:
                    await orp.write(0, 0, 0, CommandVerbs.GET_DATA, 40)
                    k, version = await orp.read_one()
                    if k == 40:
                        return version
    except Exception as e:
        raise OpenRoverException(f'Did not respond to request for OpenRover version') from e



from .openrover_protocol import OpenRoverProtocol
from .util import OpenRoverException

name = 'openrover'

# rover.OpenRover is not yet ready for primetime
__all__ = ['OpenRoverException', 'OpenRoverProtocol']



import trio

from openrover.find_device import open_rover_device
from openrover.openrover_data import OpenRoverFirmwareVersion
from openrover.openrover_protocol import CommandVerbs, OpenRoverProtocol

n = 100


async def test_protocol_write_read_immediate():
    n_received = 0
    async with open_rover_device() as device:
        protocol = OpenRoverProtocol(device)
        for i in range(n):
            with trio.fail_after(1):
                await protocol.write(0, 0, 0, CommandVerbs.GET_DATA, 40)
                key, version = await protocol.read_one()
                assert key == 40
                assert isinstance(version, OpenRoverFirmwareVersion)
                assert isinstance(version.value, int)
                assert 0 < version.value
                n_received += 1

    print(f'success ratio {n_received / n}')
    assert 0.9 < n_received / n <= 1


async def test_protocol_writes_then_reads():
    n_received = 0

    async with open_rover_device() as device:
        protocol = OpenRoverProtocol(device)

        async with trio.open_nursery() as nursery:
            for i in range(n):
                nursery.start_soon(protocol.write, 0, 0, 0, CommandVerbs.GET_DATA, 40)
            for i in range(n):
                try:
                    with trio.fail_after(1):
                        key, version = await protocol.read_one()
                        assert key == 40
                        assert isinstance(version, OpenRoverFirmwareVersion)
                        assert isinstance(version.value, int)
                        assert 0 < version.value
                        n_received += 1
                except trio.TooSlowError:
                    pass

        print(f'success ratio {n_received / n}')
        assert 0.9 < n_received / n <= 1



from math import isclose
import statistics

import pytest
import trio

from openrover import OpenRoverException
from openrover.openrover_data import OpenRoverFirmwareVersion, fix_encoder_delta
from openrover.rover import Rover, open_rover
from openrover.util import RoverDeviceNotFound


@pytest.fixture
async def rover():
    try:
        async with open_rover() as r:
            yield r
    except RoverDeviceNotFound:
        pytest.skip('This test requires a rover device but none was found')


async def test_find_openrover(rover):
    assert rover is not None
    assert isinstance(rover, Rover)


async def test_get_version(rover):
    version = await rover.get_data(40)
    assert isinstance(version, OpenRoverFirmwareVersion)

    version2 = await rover.get_data(40)
    assert version == version2
    assert 0 <= version.major <= 100
    assert 0 <= version.minor <= 100
    assert 0 <= version.patch <= 100


async def test_recover_from_bad_data(rover):
    await rover._rover_protocol._serial.write(b'test' * 20)

    for i in range(3):
        try:
            result = await rover.get_data(40)
            if result is not None:
                return
        except Exception as e:
            pass
    assert False


async def test_missing_device():
    with pytest.raises(OpenRoverException):
        async with open_rover('missing_device'):
            pass


async def test_fan_speed(rover):
    for i in range(10):
        fan_speed_cmd = i / 10
        await rover.set_fan_speed(fan_speed_cmd)
        await trio.sleep(0.1)
        fan_speed_fb = await rover.get_data(48)
        assert isclose(fan_speed_cmd, fan_speed_fb, abs_tol=0.03)


@pytest.mark.parametrize('motor_effort', [0, -0.1, +0.1, -0.2, +0.2])
async def test_encoder_intervals(rover, motor_effort):
    enc_counts_left = []
    enc_counts_right = []
    enc_intervals_left = []
    enc_intervals_right = []

    rover.set_motor_speeds(motor_effort, motor_effort, 0)

    for i in range(3):
        await rover.send_speed()
        await trio.sleep(0.1)

    for i in range(10):
        await rover.send_speed()
        await trio.sleep(0.1)
        data = await rover.get_data_items([14, 16, 28, 30])
        enc_counts_left.append(data[14])
        enc_counts_right.append(data[16])
        enc_intervals_left.append(data[28])
        enc_intervals_right.append(data[30])

    encoder_delta_left = [fix_encoder_delta(a - b) for a, b in zip(enc_counts_left[1:], enc_counts_left)]
    encoder_delta_right = [fix_encoder_delta(a - b) for a, b in zip(enc_counts_right[1:], enc_counts_right)]

    if motor_effort == 0:
        assert all(i == 0 for i in encoder_delta_left)
        assert all(i == 0 for i in encoder_delta_right)

        assert all(i == 0 or i > 500 for i in enc_intervals_left)
        assert all(i == 0 or i > 500 for i in enc_intervals_right)

    else:
        assert all(20 < i / motor_effort < 500 for i in encoder_delta_left)
        assert all(20 < i / motor_effort < 500 for i in encoder_delta_right)

        assert 0.005 < (1 / statistics.mean(enc_intervals_left)) / abs(motor_effort) < 0.05
        assert 0.005 < (1 / statistics.mean(enc_intervals_right)) / abs(motor_effort) < 0.05


async def test_currents(rover):
    battery_current_a = await rover.get_data(42)
    battery_current_b = await rover.get_data(44)
    battery_current_a_i2c = await rover.get_data(68)
    battery_current_b_i2c = await rover.get_data(70)

    # must agree within 5% or 200mA
    assert isclose(battery_current_a, abs(battery_current_a_i2c), rel_tol=0.05, abs_tol=0.2)
    assert isclose(battery_current_b, abs(battery_current_b_i2c), rel_tol=0.05, abs_tol=0.2)

    battery_current_total = await rover.get_data(0)
    assert isclose(battery_current_a + battery_current_b, battery_current_total, rel_tol=0.05, abs_tol=0.2)


async def test_soc(rover):
    battery_soc_a = await rover.get_data(34)
    battery_soc_b = await rover.get_data(34)
    assert 0 <= battery_soc_a <= 1
    assert 0 <= battery_soc_b <= 1
    assert isclose(battery_soc_a, battery_soc_b, rel_tol=0.1, abs_tol=0.1)


async def test_currents_charging(rover):
    is_charging = await rover.get_data(38)
    assert is_charging in [True, False]

    battery_current_a_i2c = await rover.get_data(68)
    battery_current_b_i2c = await rover.get_data(70)
    if is_charging:
        # charging battery has positive current
        assert battery_current_a_i2c >= 0
        assert battery_current_b_i2c >= 0
    else:
        # discharging battery should have negative current
        assert battery_current_a_i2c <= 0
        assert battery_current_b_i2c <= 0


async def test_voltages(rover):
    battery_voltage_a = await rover.get_data(24)
    battery_voltage_a_i2c = await rover.get_data(64)
    assert 14 < battery_voltage_a_i2c < 18  # voltage should be between 14-18 V
    assert 14 < battery_voltage_a < 18
    # must agree within 5% or 50mV
    assert isclose(battery_voltage_a_i2c, battery_voltage_a, rel_tol=0.05, abs_tol=0.05)

    battery_voltage_b = await rover.get_data(26)
    battery_voltage_b_i2c = await rover.get_data(66)
    assert 14 < battery_voltage_b_i2c < 18
    assert 14 < battery_voltage_b < 18
    # must agree within 5% or 50mV
    assert isclose(battery_voltage_b_i2c, battery_voltage_b, rel_tol=0.05, abs_tol=0.05)


async def test_temperatures(rover):
    fan_temp = await rover.get_data(20)
    assert 10 < fan_temp < 55

    battery_temp_a = await rover.get_data(60)
    battery_temp_b = await rover.get_data(62)

    # Rated operating temperature check batteries are between 10 and 55 degrees C
    assert 10 < battery_temp_a < 55
    assert 10 < battery_temp_b < 55
    assert abs(battery_temp_a - battery_temp_b) < 5
